/*

*/
#include <stdio.h>
#include <stdlib.h>

#ifdef MAC
   #include <OpenCL/cl.h>
#else
   #include <CL/cl.h>
#endif

#include "simpleCL.h"  // for sclPrintErrorFlag() function

// File name and function name for the OpenCL kernel
#define PROGRAM_FILE "test_opencl_workspaces.cl"
#define KERNEL_FUNC  "workspaces"

/*
  Command line args: none
*/
int main(int argc, char*argv[])
{
   /* Host/device data structures */
   cl_platform_id platform[2]; // space for two platform data structures
   cl_device_id device;
   cl_context context;
   cl_command_queue cmdQueue;

   /* Program/kernel data structures */
   cl_program program;
   FILE *program_handle;
   char *program_buffer;
   size_t program_size;
   cl_kernel kernel;
   size_t global_work_size[3];
   size_t global_work_offset[3];

   /* CL output buffer */
   cl_mem bufferA;  // Output array on the device

   cl_uint i, j;

   // Set the size of the data buffer
   size_t datasize = sizeof(cl_int) * 64;

   // Allocate and initialize a data buffer
   cl_int* val = (cl_int*)malloc( datasize );
   for (i = 0; i < 64; i++)
   {
      val[i] = 2;
   }


   // Use this to check the output of each API call
   cl_int err;


   //-----------------------------------------------------
   // STEP 1: Discover and initialize a platform
   //-----------------------------------------------------

   // Use clGetPlatformIDs() to fill in the platform data structures.
   err = clGetPlatformIDs(0, NULL, &i);            // get number of platforms
   printf("The number of platforms is %d.\n", i);
   err = clGetPlatformIDs(2, platform, NULL);      // get 2 platform data strunctures

   if(CL_SUCCESS != err)
   {  printf("Couldn't find any platforms: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 2: Discover and initialize a GPU or CPU device
   //-----------------------------------------------------

   // Change PLATFORM_NUMBER if you want to use the other platform.
   int PLATFORM_NUMBER = 1;

   // Use clGetDeviceIDs() to fill in a device data structure.
   err = clGetDeviceIDs(platform[PLATFORM_NUMBER], CL_DEVICE_TYPE_ALL, 1, NULL, &i);
   printf("The number of devices on platform %d is %d.\n", PLATFORM_NUMBER, i);
   // use the first device on the choosen platfrom
   err = clGetDeviceIDs(platform[PLATFORM_NUMBER], CL_DEVICE_TYPE_ALL, 1, &device, NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't find any devices: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 3: Create a context
   //-----------------------------------------------------

   // Use clCreateContext() to create a context data structure
   // and associate it with the GPU device.
   context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create a context: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 4: Create a command queue for the GPU device
   //-----------------------------------------------------

   // Use clCreateCommandQueue() to create a command queue
   // and associate it with the GPU device.
   cmdQueue = clCreateCommandQueue(context, device, 0, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the command queue: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 5: Create and compile the program
   //-----------------------------------------------------

   // Read program file and place content into a string buffer
   program_handle = fopen(PROGRAM_FILE, "r");
   if(NULL == program_handle)
   {  printf("Couldn't find the kernel program file: %s", PROGRAM_FILE);
      exit(1);
   }
   fseek(program_handle, 0, SEEK_END);
   program_size = ftell(program_handle);
   rewind(program_handle);
   program_buffer = (char*)malloc(program_size + 1);
   if (NULL == program_buffer)
   {  perror("malloc");
      exit(-1);
   }
   program_buffer[program_size] = '\0';
   fread(program_buffer, sizeof(char), program_size, program_handle);
   fclose(program_handle);

   // Use clCreateProgramWithSource() to create a program data structure.
   program = clCreateProgramWithSource(
       context,
       1,
       (const char**)&program_buffer,
       &program_size,
       &err);

   free(program_buffer);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the program: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clBuildProgram() to build (compile) a program for the GPU device.
   err = clBuildProgram(program,
                              0,
                           NULL,
                           "",     // compiler options
                           NULL,
                           NULL);

   if(CL_SUCCESS != err)
   {  /* Find size of log and print to std output */
      size_t log_size;
      clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG,
            0, NULL, &log_size);
      char *program_log = (char*) malloc(log_size + 1);
      program_log[log_size] = '\0';
      clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG,
            log_size + 1, program_log, NULL);
      printf("Could not compile the kernel program: ");
      sclPrintErrorFlag( err );
      printf("Build Log: ");
      printf("%s\n", program_log);
      free(program_log);
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 6: Create the kernel
   //-----------------------------------------------------

   // Use clCreateKernel() to create a kernel from the 2d_sor function.
   kernel = clCreateKernel(program, KERNEL_FUNC, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the kernel: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 7: Create device buffer to hold output data.
   //-----------------------------------------------------

   // Use clCreateBuffer() to create a CL buffer object
   // with enough space to hold the data.
   bufferA = clCreateBuffer(
       context,
       CL_MEM_WRITE_ONLY,
       datasize,
       NULL,
       &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create a buffer object: ");
      sclPrintErrorFlag( err );
      exit(1);
   }


   //-----------------------------------------------------
   // STEP 9: Create kernel argument from the CL buffer
   //-----------------------------------------------------

   // Use clSetKernelArg() to associate input and output buffers with the kernel
   err  = clSetKernelArg(
       kernel,
       0,
       sizeof(cl_mem),
       &bufferA);

   if(CL_SUCCESS != err)
   {  printf("Couldn't set the kernel arguments: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 10: Enqueue the kernel for execution
   //-----------------------------------------------------

   // The work items are in a 1, 2, or 3-dimensional work space and
   // there are (respectively) 64, 8, or 4 "work-items" in each dimension.
   global_work_size[0] = 8;  // 64 for 1-dim,  8 for 2-dim,  4 for 3-dim
   global_work_size[1] = 8;  //  8 for 2-dim,  4 for 3-dim
// global_work_size[2] = 4;

   global_work_offset[0] = 0;  // experiment with "offsets" in the workspaces
   global_work_offset[1] = 0;
   global_work_offset[2] = 0;

   // Use clEnqueueNDRangeKernel() to execute the kernel.
   err = clEnqueueNDRangeKernel(
       cmdQueue,
       kernel,
       2,        // diension of the work space (either 1, 2, or 3)
       global_work_offset,
       global_work_size,
       NULL,
       0,
       NULL,
       NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't enqueue the kernel execution command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }


   clFinish( cmdQueue );


   //-----------------------------------------------------
   // STEP 11: Read the output data back to the host
   //-----------------------------------------------------

   // Use clEnqueueReadBuffer() to read the OpenCL output
   // buffer (bufferB) to the host output array.
   clEnqueueReadBuffer(
       cmdQueue,
       bufferA,
       CL_TRUE,
       0,
       datasize,
       val,       // use the "val" array for the output
       0,
       NULL,
       NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't enqueue the kernel execution command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 12: Release OpenCL resources
   //-----------------------------------------------------

   clReleaseMemObject(bufferA);
   clReleaseKernel(kernel);
   clReleaseProgram(program);
   clReleaseCommandQueue(cmdQueue);
   clReleaseContext(context);


   for (i = 0; i < 8; i++)
   {
      for (j = 0; j < 8; j++)
      {
         printf("%d  ", val[i*8+j]);
      }
      printf("\n");
   }


   return 0;
}/* main() */
