/*
    http://www.heterogeneouscompute.org/?page_id=7
*/

// This program implements a vector addition using OpenCL

#include <stdio.h>
#include <stdlib.h>

#ifdef MAC
   #include <OpenCL/cl.h>
#else
   #include <CL/cl.h>
#endif

#include "simpleCL.h"  // for sclPrintErrorFlag() function

// OpenCL kernel to perform an element-wise add of two arrays
const char* programSource = "\
__kernel                                        \n\
void vecadd(__global int *A,                    \n\
            __global int *B,                    \n\
            __global int *C)                    \n\
{                                               \n\
   // Get the work-items unique ID             \n\
   int idx = get_global_id(0);                  \n\
                                                \n\
   // Add the corresponding locations of        \n\
   // 'A' and 'B', and store the result in 'C'. \n\
   C[idx] = 0.0; // why is this line needed?????\n\
   C[idx] = A[idx] + B[idx];                    \n\
}";

int main() // This code executes on the OpenCL host
{
   // Host data
   int *A = NULL;  // Input array
   int *B = NULL;  // Input array
   int *C = NULL;  // Output array

   // Elements in each array
   const int elements = 2048;

   // Compute the size of the data
   size_t datasize = sizeof(int)*elements;

   // Allocate space for input/output data
   A = (int*)malloc(datasize);
   B = (int*)malloc(datasize);
   C = (int*)malloc(datasize);
   // Initialize the input data
   for(int i = 0; i < elements; i++)
   {
       A[i] = i;
       B[i] = i;
   }

   // Use this to check the output of each API call
   cl_int err;

   //-----------------------------------------------------
   // STEP 1: Discover and initialize the platforms
   //-----------------------------------------------------

   cl_uint numPlatforms = 0;
   cl_platform_id *platforms = NULL;

   // Use clGetPlatformIDs() to retrieve the number of platforms
   err = clGetPlatformIDs(0, NULL, &numPlatforms);
   if(err != CL_SUCCESS)
   {  printf("Couldn't find any platforms: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Allocate enough space for each platform
   platforms = (cl_platform_id*)malloc(numPlatforms*sizeof(cl_platform_id));

   // Fill in platforms with clGetPlatformIDs()
   err = clGetPlatformIDs(numPlatforms, platforms, NULL);

   //-----------------------------------------------------
   // STEP 2: Discover and initialize the devices
   //-----------------------------------------------------

   cl_uint numDevices = 0;
   cl_device_id *devices = NULL;

   // Use clGetDeviceIDs() to retrieve the number of devices present
   err = clGetDeviceIDs(
       platforms[0],
       CL_DEVICE_TYPE_ALL,
       0,
       NULL,
       &numDevices);
   if(err != CL_SUCCESS)
   {  printf("Couldn't find any devices: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Allocate enough space for each device
   devices = (cl_device_id*)malloc(numDevices*sizeof(cl_device_id));

   // Fill in devices with clGetDeviceIDs()
   err = clGetDeviceIDs(
       platforms[0],
       CL_DEVICE_TYPE_ALL,
       numDevices,
       devices,
       NULL);

   //-----------------------------------------------------
   // STEP 3: Create a context
   //-----------------------------------------------------

   cl_context context = NULL;

   // Create a context using clCreateContext() and
   // associate it with the devices
   context = clCreateContext(
       NULL,
       numDevices,
       devices,
       NULL,
       NULL,
       &err);
   if(err != CL_SUCCESS)
   {  printf("Couldn't create a context: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 4: Create a command queue
   //-----------------------------------------------------

   cl_command_queue cmdQueue;

   // Create a command queue using clCreateCommandQueue(),
   // and associate it with the device you want to execute on.
   cmdQueue = clCreateCommandQueue(
       context,
       devices[0],
       0,
       &err);
   if(err != CL_SUCCESS)
   {  printf("Couldn't create the command queue: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 5: Create and compile the program
   //-----------------------------------------------------

   // Create a program using clCreateProgramWithSource()
   cl_program program = clCreateProgramWithSource(
       context,
       1,
       (const char**)&programSource,
       NULL,
       &err);
   if(err != CL_SUCCESS)
   {  printf("Couldn't create the program: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Build (compile) the program for the devices with clBuildProgram().
   err = clBuildProgram(
       program,
       numDevices,
       devices,
       NULL,
       NULL,
       NULL);
   if(err != CL_SUCCESS)
   {  int i;
      for (i=0; i<numDevices; i++)
      {  /* Find size of log and print to std output */
         size_t log_size;
         clGetProgramBuildInfo(program, devices[i], CL_PROGRAM_BUILD_LOG,
               0, NULL, &log_size);
         char *program_log = (char*) malloc(log_size + 1);
         program_log[log_size] = '\0';
         clGetProgramBuildInfo(program, devices[i], CL_PROGRAM_BUILD_LOG,
               log_size + 1, program_log, NULL);
         printf("Could not compile the kernel program for device %d: ", i);
         sclPrintErrorFlag( err );
         printf("Build Log: ");
         printf("%s\n", program_log);
         free(program_log);
      }
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 6: Create the kernel
   //-----------------------------------------------------

   cl_kernel kernel = NULL;

   // Use clCreateKernel() to create a kernel from the
   // vector addition function (named "vecadd").
   kernel = clCreateKernel(program, "vecadd", &err);
   if(err != CL_SUCCESS)
   {  printf("Couldn't create the kernel: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 7: Create device buffers
   //-----------------------------------------------------

   cl_mem bufferA;  // Input array on the device
   cl_mem bufferB;  // Input array on the device
   cl_mem bufferC;  // Output array on the device

   // Use clCreateBuffer() to create a buffer object (d_A)
   // that will contain the data from the host array A.
   bufferA = clCreateBuffer(
       context,
       CL_MEM_READ_ONLY,
       datasize,
       NULL,
       &err);
   if(err != CL_SUCCESS)
   {  printf("Couldn't create a buffer object: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clCreateBuffer() to create a buffer object (d_B)
   // that will contain the data from the host array B.
   bufferB = clCreateBuffer(
       context,
       CL_MEM_READ_ONLY,
       datasize,
       NULL,
       &err);

   // Use clCreateBuffer() to create a buffer object (d_C)
   // with enough space to hold the output data.
   bufferC = clCreateBuffer(
       context,
       CL_MEM_WRITE_ONLY,
       datasize,
       NULL,
       &err);

   //-----------------------------------------------------
   // STEP 8: Write host data to device buffers
   //-----------------------------------------------------

   // Use clEnqueueWriteBuffer() to write input array A to
   // the device buffer bufferA.
   err = clEnqueueWriteBuffer(
       cmdQueue,
       bufferA,
       CL_FALSE,
       0,
       datasize,
       A,
       0,
       NULL,
       NULL);
   if(err != CL_SUCCESS)
   {  printf("Couldn't enqueue the write buffer command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clEnqueueWriteBuffer() to write input array B to
   // the device buffer bufferB.
   err = clEnqueueWriteBuffer(
       cmdQueue,
       bufferB,
       CL_FALSE,
       0,
       datasize,
       B,
       0,
       NULL,
       NULL);
   if(err != CL_SUCCESS)
   {  printf("Couldn't enqueue the write buffer command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 9: Set the kernel arguments
   //-----------------------------------------------------

   // Associate the input and output buffers with the
   // kernel using clSetKernelArg()
   err  = clSetKernelArg(
       kernel,
       0,
       sizeof(cl_mem),
       &bufferA);
   err |= clSetKernelArg(
       kernel,
       1,
       sizeof(cl_mem),
       &bufferB);
   err |= clSetKernelArg(
       kernel,
       2,
       sizeof(cl_mem),
       &bufferC);
   if(err != CL_SUCCESS)
   {  printf("Couldn't set the kernel arguments: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 10: Configure the work-item structure
   //-----------------------------------------------------

   // Define an index space (global work size) of work items
   // for execution. A workgroup size (local work size)
   // is not required, but can be used.
   size_t globalWorkSize[1];
   // There are 'elements' work-items
   globalWorkSize[0] = elements;


   //-----------------------------------------------------
   // STEP 11: Enqueue the kernel for execution
   //-----------------------------------------------------

   // Execute the kernel by using clEnqueueNDRangeKernel().
   // 'globalWorkSize' is the 1D dimension of the work-items.
   err = clEnqueueNDRangeKernel(
       cmdQueue,
       kernel,
       1,
       NULL,
       globalWorkSize,
       NULL,
       0,
       NULL,
       NULL);
   if(err != CL_SUCCESS)
   {  printf("Couldn't enqueue the kernel execution command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 12: Read the output buffer back to the host
   //-----------------------------------------------------

   // Use clEnqueueReadBuffer() to read the OpenCL output
   // buffer (bufferC) to the host output array (C)
   err = clEnqueueReadBuffer(
       cmdQueue,
       bufferC,
       CL_TRUE,
       0,
       datasize,
       C,
       0,
       NULL,
       NULL);
   if(err != CL_SUCCESS)
   {  printf("Couldn't enqueue the read buffer command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Verify the output
   int result = 1;
   for(int i = 0; i < elements; i++)
   {
      if(C[i] != i+i)
      {
         result = 0;
         break;
      }
   }
   if(result)
   {
      printf("Output is correct\n");
   }
   else
   {
      printf("Output is incorrect\n");
   }

   //-----------------------------------------------------
   // STEP 13: Release OpenCL resources
   //-----------------------------------------------------

   // Free OpenCL resources
   clReleaseKernel(kernel);
   clReleaseProgram(program);
   clReleaseCommandQueue(cmdQueue);
   clReleaseMemObject(bufferA);
   clReleaseMemObject(bufferB);
   clReleaseMemObject(bufferC);
   clReleaseContext(context);

   // Free host resources
   free(A);
   free(B);
   free(C);
   free(platforms);
   free(devices);
}
